use std::time::Duration;

use bevy::prelude::*;

use crate::time::GameTime;

use super::{MapMarker, MeshNeedsUpdate};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum CellKind {
    Sea,
    Beach,
    Forest,
    Dirt,
    Stone,
    Grass
}
impl CellKind {
    pub fn regen_full_growth_duration(&self) -> Duration {
        match self {
            CellKind::Sea | CellKind::Beach | CellKind::Dirt | CellKind::Stone => unreachable!(),
            CellKind::Forest => Duration::from_days(365*100), // Let's say that a forest takes 100 years to mature
            CellKind::Grass => Duration::from_weeks(7) // Let's say that grass takes 7 weeks to reach its max
        }
    }
}

#[derive(Debug, Component)]
pub struct Cell {
    pub kind: CellKind,
    pub voronoi_id: usize,
    pub altitude: u8,
    pub vertices: Vec<usize>
}
impl Cell {
    // pub fn new(kind: CellKind, voronoi_id: usize, altitude: u8, moisture: u8, vertices: Vec<usize>) -> Self {
    //     Self {
    //         kind,
    //         voronoi_id,
    //         altitude,
    //         moisture,
    //         vertices
    //     }
    // }
    pub fn color(&self, wealth: u8) -> [f32; 4] {
        match self.kind {
            CellKind::Sea => [0., 0., 1., 1.],
            CellKind::Beach => [0.82, 0.84, 0.51, 1.],
            CellKind::Forest => [0., 0.5 - (wealth as f32/255.*0.4), 0., 1.],
            CellKind::Dirt => [0.53 - (wealth as f32/255.*0.4), 0.38-(wealth as f32/255.*0.4), 0.29-(wealth as f32/255.*0.4), 1.],
            CellKind::Stone => [0.5, 0.5, 0.5, 1.],
            CellKind::Grass => [(136./255.) - (wealth as f32/255.*0.4), (204./255.) - (wealth as f32/255.*0.4), (59./255.) - (wealth as f32/255.*0.4), 1.]
        }
    }
}

#[derive(Component)]
pub struct Wealth(pub u8);

impl Default for Wealth {
    fn default() -> Self {
        Wealth(u8::MAX)
    }
}

#[derive(Component)]
#[require(Wealth)]
pub struct Regeneration {
    pub last_update: Duration,
    pub full_growth_duration: Duration
}

pub fn cells_regeneration(
    mut cells: Query<(&mut Regeneration, &mut Wealth)>,
    mut map_needs_update: Query<&mut MeshNeedsUpdate, With<MapMarker>>,
    gt: Res<GameTime>
) {
    let mut map_needs_update = map_needs_update.single_mut();
    for (mut regen, mut wealth) in cells.iter_mut() {
        if gt.current - regen.last_update > regen.full_growth_duration/u8::MAX as u32 {
            regen.last_update = gt.current;
            wealth.0 = wealth.0.saturating_add(1);
            map_needs_update.0 = true;
        }
    }
}